
class ByteBuffer(object):
    def __init__(self):
        # 字节缓冲区
        self.buffer = bytearray([00]*1024)
        # 写入的位置索引号
        self.wpos = 0
        # 读出的位置索引号
        self.rpos = 0
        # 大小端格式
        self.endianness = "little"
        # 索引的数组及快速查询的字符串索引号
        self.str_arr = []
        self.str_map = {}

    @staticmethod
    def allocate(n):
        d = bytearray([00] * n)
        b = ByteBuffer()
        b.__init(d)
        return b

    @staticmethod
    def wrap(data):
        b = ByteBuffer()
        b.__init(data)
        return b

    def __init(self, data):
        if type(data) is not bytearray:
            data = bytearray(data)

        self.buffer = data
        self.wpos = 0
        self.rpos = 0

    def _check_buffer(self, space):
        return (self.wpos - self.rpos) >= space
    
    def _ensure_buffer(self, space):
        if len(self.buffer) - self.wpos > space:
            return True
        now_len = len(self.buffer)
        if now_len < space:
            now_len += space
        new_data = bytearray([00] * now_len)
        self.buffer.extend(new_data)
        return True

    def _read(self, length=1):
        assert self._check_buffer(length), 'Buffer has not enough bytes to read'
        r = self.buffer[self.rpos:self.rpos + length]
        self._read_offsets(length)
        return r

    def _write_offsets(self, value):
        self.wpos += value
        
    def _read_offsets(self, value):
        self.rpos += value
        if self.rpos == self.wpos:
            self.rpos = 0
            self.wpos = 0

    def read(self, length=1):
        return int.from_bytes(self._read(length=length), byteorder=self.endianness)

    def write(self, value, size=0, signed=False):
        from tunm_proto import TP_DATA_TYPE
        t = type(value)
        if t == int:
            b = int.to_bytes(value, size, byteorder=self.endianness, signed=signed)
        elif t == TP_DATA_TYPE:
            b = int.to_bytes(int(value), 1, byteorder=self.endianness)
        elif t == list:
            b = bytes(value)
        elif t == bytes or t == bytearray:
            b = value
        else:
            raise Exception('Attempting to write unknown object into Buffer')
        l = len(b)
        assert self._ensure_buffer(l), 'Buffer has not enough space left'
        for i in range(l):
            self.buffer[self.wpos + i:self.wpos + i + 1] = b[i:i + 1]
        self._write_offsets(l)
        
    def write_u8(self, value):
        self.write(value, 1)
        
    def write_i8(self, value):
        self.write(value, 1, True)
        
    def write_u16(self, value):
        self.write(value, 2)
        
    def write_i16(self, value):
        self.write(value, 2)
        
    def write_u32(self, value):
        self.write(value, 4)
        
    def write_i32(self, value):
        self.write(value, 4)
        
    def write_u64(self, value):
        self.write(value, 8)
        
    def write_i64(self, value):
        self.write(value, 8)
        
    def read_u8(self):
        return self.read(1)
        
    def read_i8(self):
        return self.read(1)
        
    def read_u16(self):
        return self.read(2)
        
    def read_i16(self):
        return self.read(2)
        
    def read_u32(self):
        return self.read(4)
        
    def read_i32(self):
        return self.read(4)
        
    def read_u64(self):
        return self.read(8)
        
    def read_i64(self):
        return self.read(8)
    
    def write_str(self, s: str):
        return self.write(s.encode("utf-8"))
    
    def read_str(self, length):
        r = self._read(length)
        return r.decode("utf-8")
    
    def read_bytes(self, length):
        return self._read(length)
    
    def write_bytes(self, b: bytes):
        return self.write(b)
    
    def add_str(self, val):
        if val in self.str_map:
            return self.str_map[val]
        else:
            self.str_arr.append(val)
            self.str_map[val] = len(self.str_arr) - 1
            return len(self.str_arr) - 1
    
    def get_str(self, idx):
        if idx < 0 or idx > len(self.str_arr):
            raise Exception("out of range")
        else:
            return self.str_arr[idx]

    def rewind(self):
        self.wpos = 0
        self.rpos = 0
        
    def all_bytes(self):
        return self.buffer[self.rpos:self.wpos]

    def get_bytes_len(self):
        return self.wpos - self.rpos

    



from tunm_proto import ByteBuffer, TPPacker, TP_DATA_TYPE

def func(x):
    return x +1

def test_answer():
    assert func(3)==4
    
def check_buffer_data(buffer: ByteBuffer, bs):
    mbs = buffer.all_bytes()
    assert buffer.get_bytes_len() == len(bs), "length must equal"
    for i, d in enumerate(bs):
        assert int.from_bytes(mbs[i:i+1], byteorder="little") == d, "data must equal"
    
    
    
def test_encode_u8():
    buffer = ByteBuffer();

    TPPacker.encode_field(buffer, 12)
    check_buffer_data(buffer, [2, 12])
    
    number = TPPacker.decode_field(buffer)
    assert number == 12, "Number Not Match"

    
def test_encode_16():
    buffer = ByteBuffer();

    TPPacker.encode_field(buffer, -1, TP_DATA_TYPE.TYPE_I16)
    TPPacker.encode_field(buffer, 0x1234, TP_DATA_TYPE.TYPE_U16)
    check_buffer_data(buffer, [10, 1, 10, 232, 72])
    
    number = TPPacker.decode_field(buffer)
    assert number == -1, "Number Not Match"
    number = TPPacker.decode_field(buffer)
    assert number == 0x1234, "Number Not Match"

def test_encode_u32():
    buffer = ByteBuffer();

    TPPacker.encode_field(buffer, -90, TP_DATA_TYPE.TYPE_I32)
    TPPacker.encode_field(buffer, 0x12345678, TP_DATA_TYPE.TYPE_U32)
    check_buffer_data(buffer, [10, 179, 1, 10, 240, 217, 162, 163, 2])
    
    number = TPPacker.decode_field(buffer)
    assert number == -90, "Number Not Match"
    number = TPPacker.decode_field(buffer)
    assert number == 0x12345678, "Number Not Match"
    
    
def test_encode_float():
    buffer = ByteBuffer();

    TPPacker.encode_field(buffer, 12345.123, TP_DATA_TYPE.TYPE_FLOAT)
    check_buffer_data(buffer, [11, 35, 95, 188, 0])
    
    number = TPPacker.decode_field(buffer)
    assert number == 12345.123, "Number Not Match"
    
    
def test_encode_str():
    buffer = ByteBuffer();

    name = "this is tunm proto, 中文测试";
    TPPacker.encode_str_raw(buffer, name, TP_DATA_TYPE.TYPE_STR);
    
    check_buffer_data(buffer, [64, 116, 104, 105, 115, 32, 105, 115, 32, 116, 117, 110, 109, 32, 112, 114, 111, 116, 111, 44, 32, 228, 184, 173, 230, 150, 135, 230, 181, 139, 232, 175, 149])
    
    read = TPPacker.decode_str_raw(buffer, TP_DATA_TYPE.TYPE_STR)
    assert read == name, "str Not Match"
    
def test_encode_map():
    buffer = ByteBuffer();

    hash_value = {
        "name": "tickbh",
        "proto": "tunm",
        "index": 1,
    }
    TPPacker.encode_field(buffer, hash_value)
    read = TPPacker.decode_field(buffer)
    assert read == hash_value, "map Not Match"
    
    
def test_encode_array():
    buffer = ByteBuffer();
    array = []
    
    for i in range(10):
        array.append(i);
    TPPacker.encode_field(buffer, array)
    read = TPPacker.decode_field(buffer)
    assert read == array, "array Not Match"
    
def test_base_proto():
    buffer = ByteBuffer();
    array = []
    
    hash_value = {
        "name": "tickbh",
        "proto": "tunm",
        "index": 1,
    }
    
    TPPacker.encode_proto(buffer, "cmd_test_op", [hash_value])
    name, read = TPPacker.decode_proto(buffer)
    assert name == "cmd_test_op", "array Not Match"
    assert read == [hash_value], "array Not Match"
    


import enum
from enum import IntEnum
from math import floor

from tunm_proto import ByteBuffer
 
@enum.unique
class TP_DATA_TYPE(IntEnum):
    TYPE_NIL = 0,
    TYPE_BOOL = 1,
    TYPE_U8 = 2,
    TYPE_I8 = 3,
    TYPE_U16 = 4,
    TYPE_I16 = 5,
    TYPE_U32 = 6,
    TYPE_I32 = 7,
    TYPE_U64 = 8,
    TYPE_I64 = 9,
    TYPE_VARINT = 10,
    TYPE_FLOAT = 11,
    TYPE_DOUBLE = 12,
    TYPE_STR = 13,
    TYPE_STR_IDX = 14,
    TYPE_RAW = 15,
    TYPE_ARR = 16,
    TYPE_MAP = 17,
    
class TPPacker:
    @staticmethod
    def get_type_by_ref(ref_type):
        t = type(ref_type)
        if t == str:
            return TP_DATA_TYPE.TYPE_STR
        elif t == bytes:
            return TP_DATA_TYPE.TYPE_RAW
        elif t == dict:
            return TP_DATA_TYPE.TYPE_MAP
        elif t == list:
            return TP_DATA_TYPE.TYPE_ARR
        elif t == int:
            if ref_type < 256 and ref_type >= 0:
                return TP_DATA_TYPE.TYPE_U8
            elif ref_type < 128 and ref_type >= -128:
                return TP_DATA_TYPE.TYPE_I8
            return TP_DATA_TYPE.TYPE_I64
        elif t == float:
            return TP_DATA_TYPE.TYPE_DOUBLE
        return TP_DATA_TYPE.TYPE_NIL
    
    @staticmethod        
    def decode_varint(buffer: ByteBuffer):
        real = 0
        shl_num = 0
        while True:
            data = buffer.read_u8()
            real += (data & 0x7F) << shl_num
            shl_num += 7
            if (data & 0x80) == 0:
                break
        is_left = real % 2 == 1
        if is_left:
            return -floor(real / 2) - 1
        else:
            return floor(real / 2)
        
    @staticmethod        
    def decode_type(buffer: ByteBuffer):
        return buffer.read_u8()
    
    
    @staticmethod        
    def decode_number(buffer: ByteBuffer, pattern):
        if pattern == TP_DATA_TYPE.TYPE_U8:
            return buffer.read_u8()
        elif pattern == TP_DATA_TYPE.TYPE_I8:
            return buffer.read_i8()
        elif pattern == TP_DATA_TYPE.TYPE_U16:
            return buffer.read_u16()
        elif pattern == TP_DATA_TYPE.TYPE_I16:
            return buffer.read_i16()
        elif pattern == TP_DATA_TYPE.TYPE_U32:
            return buffer.read_u32()
        elif pattern == TP_DATA_TYPE.TYPE_I32:
            return buffer.read_i32()
        elif pattern == TP_DATA_TYPE.TYPE_U64:
            return buffer.read_u64()
        elif pattern == TP_DATA_TYPE.TYPE_I64:
            return buffer.read_i64()
        elif pattern == TP_DATA_TYPE.TYPE_FLOAT:
            return buffer.read_i32() / 1000.0
        elif pattern == TP_DATA_TYPE.TYPE_DOUBLE:
            return buffer.read_i64() / 1000000.0
        else:
            raise Exception(f"Unknow decode number type {pattern}")
        
    @staticmethod        
    def decode_str_raw(buffer: ByteBuffer, pattern):
        if pattern == TP_DATA_TYPE.TYPE_STR:
            length = TPPacker.decode_varint(buffer)
            if length == 0:
                return ""
            return buffer.read_str(length)
        elif pattern == TP_DATA_TYPE.TYPE_RAW:
            length = TPPacker.decode_varint(buffer)
            if length == 0:
                return ""
            return buffer.read_bytes(length)
        else:
            raise Exception(f"Unknow decode str type {pattern}")
    
    
    @staticmethod        
    def decode_field(buffer: ByteBuffer):
        pattern = TPPacker.decode_type(buffer)
        if pattern == TP_DATA_TYPE.TYPE_NIL:
            return None
        elif pattern == TP_DATA_TYPE.TYPE_BOOL:
            return True if buffer.read_u8() != 0 else False
        elif pattern >= TP_DATA_TYPE.TYPE_U8 and pattern <= TP_DATA_TYPE.TYPE_I8:
            return TPPacker.decode_number(buffer, pattern)
        elif pattern >= TP_DATA_TYPE.TYPE_U16 and pattern <= TP_DATA_TYPE.TYPE_I64:
            return TPPacker.decode_varint(buffer)
        elif pattern == TP_DATA_TYPE.TYPE_FLOAT:
            return TPPacker.decode_number(buffer, TP_DATA_TYPE.TYPE_I32) / 1000.0
        elif pattern == TP_DATA_TYPE.TYPE_DOUBLE:
            return TPPacker.decode_number(buffer, TP_DATA_TYPE.TYPE_I64) / 1000000.0
        elif pattern == TP_DATA_TYPE.TYPE_VARINT:
            return TPPacker.decode_varint(buffer) 
        elif pattern == TP_DATA_TYPE.TYPE_STR_IDX:
            idx = TPPacker.decode_varint(buffer)
            return buffer.get_str(idx)
        elif pattern == TP_DATA_TYPE.TYPE_STR or pattern == TP_DATA_TYPE.TYPE_RAW:
            return TPPacker.decode_str_raw(buffer, pattern)
        elif pattern == TP_DATA_TYPE.TYPE_ARR:
            return TPPacker.decode_arr(buffer)
        elif pattern == TP_DATA_TYPE.TYPE_MAP:
            return TPPacker.decode_map(buffer)
        else:
            raise Exception("unknow type")
        
    
    
    @staticmethod        
    def decode_arr(buffer: ByteBuffer):
        size = TPPacker.decode_varint(buffer)
        arr = []
        for _idx in range(size):
            sub = TPPacker.decode_field(buffer)
            arr.append(sub)
        return arr
    
    
    @staticmethod        
    def decode_map(buffer: ByteBuffer):
        size = TPPacker.decode_varint(buffer)
        map = {}
        for _idx in range(size):
            key = TPPacker.decode_field(buffer)
            val = TPPacker.decode_field(buffer)
            map[key] = val
        return map
    
    @staticmethod        
    def decode_proto(buffer: ByteBuffer):
        name = TPPacker.decode_str_raw(buffer, TP_DATA_TYPE.TYPE_STR)
        str_len = TPPacker.decode_varint(buffer)
        for _ in range(str_len):
            value = TPPacker.decode_str_raw(buffer, TP_DATA_TYPE.TYPE_STR)
            buffer.add_str(value)
        sub_value = TPPacker.decode_field(buffer)
        return name, sub_value
    
    
    @staticmethod
    def encode_varint(buffer: ByteBuffer, value):
        '''
        如果原数值是正数则将原数值变成value*2
        如果原数值是负数则将原数值变成-(value + 1) * 2 + 1
        相当于0->0, -1->1, 1->2,-2->3,2->4来做处理
        因为小数值是常用的, 所以保证小数值及负数的小数值尽可能的占少位
        '''
        if type(value) == bool:
            value = 1 if value else 0
        real = value * 2
        if value < 0:
            real = -(value + 1) * 2 + 1
        
        for _i in range(12):
            # 每个字节的最高位来表示有没有下一位, 若最高位为0, 则已完毕
            b = real & 0x7F
            real >>= 7
            if real > 0:
                buffer.write_u8(b | 0x80)
            else:
                buffer.write_u8(b)
                break
        
    @staticmethod
    def encode_type(buffer: ByteBuffer, value):
        buffer.write_u8(value)
        
    @staticmethod
    def encode_bool(buffer: ByteBuffer, value):
        buffer.write_u8(1 if value else 0)
        
    @staticmethod
    def encode_number(buffer: ByteBuffer, value, pattern):
        if pattern == TP_DATA_TYPE.TYPE_U8:
            return buffer.write_u8(value)
        elif pattern == TP_DATA_TYPE.TYPE_I8:
            return buffer.write_i8(value)
        elif pattern == TP_DATA_TYPE.TYPE_U16:
            return buffer.write_u16(value)
        elif pattern == TP_DATA_TYPE.TYPE_I16:
            return buffer.write_i16(value)
        elif pattern == TP_DATA_TYPE.TYPE_U32:
            return buffer.write_u32(value)
        elif pattern == TP_DATA_TYPE.TYPE_I32:
            return buffer.write_i32(value)
        elif pattern == TP_DATA_TYPE.TYPE_U64:
            return buffer.write_u64(value)
        elif pattern == TP_DATA_TYPE.TYPE_I64:
            return buffer.write_i64(value)
        elif pattern == TP_DATA_TYPE.TYPE_FLOAT:
            return buffer.write_i32(floor(value * 1000.0))
        elif pattern == TP_DATA_TYPE.TYPE_DOUBLE:
            return buffer.write_i64(floor(value * 1000000.0))
        else:
            raise Exception(f"Unknow decode number type {pattern}")
            
    @staticmethod
    def encode_str_idx(buffer: ByteBuffer, value):
        '''
        写入字符串索引值, 在数值区里的所有字符串默认会被写成索引值
        如果重复的字符串则会返回相同的索引值(varint)
        '''
        idx = buffer.add_str(value)
        TPPacker.encode_type(buffer, TP_DATA_TYPE.TYPE_STR_IDX)
        TPPacker.encode_varint(buffer, idx)
        
    @staticmethod
    def encode_str_raw(buffer: ByteBuffer, value, pattern):
        if pattern == TP_DATA_TYPE.TYPE_STR:
            b = value.encode("utf-8")
            TPPacker.encode_varint(buffer, len(b))
            buffer.write_bytes(b)
        elif pattern == TP_DATA_TYPE.TYPE_RAW:
            TPPacker.encode_varint(buffer, len(value))
            buffer.write_bytes(value)
        else:
            raise Exception(f"Unknow decode str type {pattern}")
        
    @staticmethod        
    def encode_field(buffer: ByteBuffer, value, pattern=None):
        '''
        先写入类型的值(u8), 则根据类型写入类型对应的的数据
        '''
        if not pattern:
            pattern = TPPacker.get_type_by_ref(value)
        if pattern == TP_DATA_TYPE.TYPE_NIL:
            return None
        elif pattern == TP_DATA_TYPE.TYPE_BOOL:
            TPPacker.encode_type(buffer, pattern)
            TPPacker.encode_bool(buffer, value)
        elif pattern >= TP_DATA_TYPE.TYPE_U8 and pattern <= TP_DATA_TYPE.TYPE_I8:
            TPPacker.encode_type(buffer, pattern)
            TPPacker.encode_number(buffer, value, pattern)
        elif pattern >= TP_DATA_TYPE.TYPE_U16 and pattern <= TP_DATA_TYPE.TYPE_I64:
            TPPacker.encode_type(buffer, TP_DATA_TYPE.TYPE_VARINT)
            TPPacker.encode_varint(buffer, value)
        elif pattern == TP_DATA_TYPE.TYPE_FLOAT:
            TPPacker.encode_type(buffer, pattern)
            TPPacker.encode_number(buffer, value, pattern)
        elif pattern == TP_DATA_TYPE.TYPE_DOUBLE:
            TPPacker.encode_type(buffer, pattern)
            TPPacker.encode_number(buffer, value, pattern)
        elif pattern == TP_DATA_TYPE.TYPE_STR:
            TPPacker.encode_str_idx(buffer, value)
        elif pattern == TP_DATA_TYPE.TYPE_RAW:
            TPPacker.encode_type(buffer, pattern)
            TPPacker.encode_str_raw(buffer, value)
        elif pattern == TP_DATA_TYPE.TYPE_ARR:
            TPPacker.encode_type(buffer, pattern)
            TPPacker.encode_arr(buffer, value)
        elif pattern == TP_DATA_TYPE.TYPE_MAP:
            TPPacker.encode_type(buffer, pattern)
            TPPacker.encode_map(buffer, value)
        else:
            raise Exception("unknow type")
        
    @staticmethod
    def encode_arr(buffer: ByteBuffer, value):
        '''
        写入数组的长度, 再写入各各元素的值
        '''
        TPPacker.encode_varint(buffer, len(value))
        for v in value:
            TPPacker.encode_field(buffer, v)
            
    @staticmethod
    def encode_map(buffer: ByteBuffer, value):
        '''
        写入map的长度, 再分别写入map各元素的key, value值
        '''
        TPPacker.encode_varint(buffer, len(value))
        for k in value:
            TPPacker.encode_field(buffer, k)
            TPPacker.encode_field(buffer, value[k])

    @staticmethod
    def encode_proto(buffer: ByteBuffer, name, infos):
        '''
        写入协议名称, 然后写入字符串索引区(即字符串数组), 然后再写入协议的详细数据
        '''
        sub_buffer = ByteBuffer()
        TPPacker.encode_field(sub_buffer, infos)

        TPPacker.encode_str_raw(buffer, name, TP_DATA_TYPE.TYPE_STR)
        TPPacker.encode_varint(buffer, len(sub_buffer.str_arr))
        for val in sub_buffer.str_arr:
            TPPacker.encode_str_raw(buffer, val, TP_DATA_TYPE.TYPE_STR)

        buffer.write_bytes(sub_buffer.all_bytes())